/* midi - MIDI file library and Simple Software Synthesizer
 * Copyright (C) 2004 by Patrick Reynolds <reynolds@cs.duke.edu>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */

#include <fcntl.h>
#include <getopt.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/soundcard.h>
#include <sys/ioctl.h>
#include "midifile.h"

#ifndef MIN
#define MIN(a,b) (((a)<(b))?(a):(b))
#endif

#define DSP "/dev/dsp"
#define DSP_SPEED 44100
#define DSP_SAMPLESIZE 16
#define CHECK(msg,n) {if (n==-1) { perror(msg); exit(1); } }

static void usage(const char *prog);
static void playfile(const char *filename);
static int init(void);
static void fill(signed short *buf, double hz, int len, int age, int ampl);
static void flushem(int samples, int *vel, int *age, int fd);

static int piano_roll = 0;
static float decay_rate = 1;
enum { WAVE_TRIANGLE, WAVE_SQUARE, WAVE_SINE } wave_type = WAVE_TRIANGLE;

int main(int argc, char **argv) {
	int i;
	char c;

	while ((c = getopt(argc, argv, "d:pw:")) != -1) {
		switch (c) {
			case 'd':
				decay_rate = atof(optarg);
				break;
			case 'p':
				piano_roll = 1;
				break;
			case 'w':
				switch (optarg[0]) {
					case 't':  wave_type = WAVE_TRIANGLE;  break;
					case 'q':  wave_type = WAVE_SQUARE;    break;
					case 's':  wave_type = WAVE_SINE;      break;
					default:     usage(argv[0]);
				}
				break;
			default:
				usage(argv[0]);
		}
	}
	if (optind > argc-1) usage(argv[0]);
	
	for (i=optind; i<argc; i++)
		playfile(argv[i]);

	return 0;
}

static int init(void) {
	int fd, abuf_size, samplesize=DSP_SAMPLESIZE, dspstereo=0, dspspeed=DSP_SPEED;
	fd = open(DSP, O_WRONLY|O_CREAT|O_TRUNC, 0666);
	if (fd == -1) { perror(DSP); return 1; }

	CHECK("ioctl", ioctl(fd, SNDCTL_DSP_RESET, 0));
	CHECK("ioctl", ioctl(fd, SNDCTL_DSP_GETBLKSIZE, &abuf_size));
	CHECK("ioctl", ioctl(fd, SNDCTL_DSP_SPEED, &dspspeed));
	CHECK("ioctl", ioctl(fd, SNDCTL_DSP_SAMPLESIZE, &samplesize));
	if (samplesize != DSP_SAMPLESIZE) {
		fprintf(stderr, "Sample size: %d instead of %d\n", samplesize, DSP_SAMPLESIZE);
		exit(1);
	}
	CHECK("ioctl", ioctl(fd, SNDCTL_DSP_STEREO, &dspstereo));

	return fd;
}

static void playfile(const char *filename) {
	int fd, vel[128], age[128];
	int evnum, last_time = 0;
	int tempo = 120;
	MIDIFile *mf;
	FILE *fp;

	fp = fopen(filename, "r");
	if (!fp) {
		perror(filename);
		return;
	}
	mf = read_midi_file(fp);
	merge_tracks(mf);

	fd = init();

	bzero(vel, sizeof(vel));
	for (evnum=0; evnum<TRACK(mf, 0)->events->len-1; evnum++) {
		Event *ev;
		
		ev = EVENT(mf, 0, evnum);

		if (ev->status >= 0x80 && ev->status <= 0x9F && ev->start > last_time) {
			flushem(DSP_SPEED * 60 / tempo / mf->header->division * (ev->start - last_time), vel, age, fd);

			last_time = ev->start;
		}

		if (ev->status >= 0x90 && ev->status <= 0x9F) {
			vel[ev->pitch] = ev->velocity;
			age[ev->pitch] = 0;
		}
		else if (ev->status >= 0x80 && ev->status <= 0x8F)
			vel[ev->pitch] = 0;
		else if (ev->status == 0xFF && ev->pitch == 0x51) {
			tempo = ev->velocity;
			if (piano_roll) printf("TEMPO: %d\n", tempo);
			continue;
		}
	}

	free_midi_file(mf);

	close(fd);
}

static void fill(signed short *buf, double hz, int len, int age, int ampl) {
	int i, period = DSP_SPEED/hz;

	debug(3, "hz=%.3f len=%d ampl=%d -> period=%d\n", hz, len, ampl, period);
	for (i=0; i<len; i++) {
		int adj, newval;
		switch (wave_type) {
			case WAVE_TRIANGLE:
				if ((i+age+period/4)%period < period/2) {
					adj = - ampl/2 + (2*ampl*((i+age+period/4)%period))/period;
				}
				else {
					adj = - ampl/2 + (2*ampl*(period-(i+age+period/4)%period))/period;
				}
				break;

			case WAVE_SQUARE:
				if ((i+age)%period < period/2) {
					adj = ampl/2;
				}
				else {
					adj = - ampl/2;
				}
				break;

			case WAVE_SINE:
				adj = ampl*sin((age+i)*2*M_PI/period);
				break;
		}

		/* decay is exponental; rate is linear with pitch */
		/* middle C (262Hz) decays e^{-1} every second */
		if (decay_rate != 0.0)
			adj *= exp(-decay_rate*(i+age)*(hz/262)/DSP_SPEED);

		newval = buf[i] + adj;
		if (newval > 32767) buf[i] = 32767;
		else if (newval < -32768) buf[i] = -32768;
		else buf[i] = newval;
	}
}

static void flushem(int samples, int *vel, int *age, int fd) {
	int i;
	signed short buf[1048576];

	bzero(buf, samples*2);
	for (i=0; i<128; i++)
		if (vel[i]) {
			fill(buf, 8.17579892 * pow(2, i/12.), samples, age[i], 5000*vel[i]/127);
			age[i] += samples;
		}
	//CHECK("ioctl", ioctl(fd, SNDCTL_DSP_SYNC, NULL));
	if (piano_roll) {
		for (i=25; i<105; i++) {
			if (vel[i] >= 96) putchar('*');
			else if (vel[i] >= 64) putchar('+');
			else if (vel[i] >= 32) putchar('-');
			else if (vel[i] >= 1) putchar('.');
			else putchar(' ');
		}
		putchar('\n');
	}
	for (i=0; i<samples*2; i+=512)
		write(fd, buf+i/2, MIN(512, samples*2-i));
}

static void usage(const char *prog) {
	fprintf(stderr, "Usage:\n  %s [options] file.mid [file2.mid [...]]\n", prog);
	fprintf(stderr, "  -d<decay>      set decay rate (default=1)\n");
	fprintf(stderr, "  -p             enable piano-roll output\n");
	fprintf(stderr, "  -w{t,s,q}      select wave form\n");
	fprintf(stderr, "                   t = triangle/sawtooth (default)\n");
	fprintf(stderr, "                   s = sine\n");
	fprintf(stderr, "                   q = square\n");
	exit(1);
}
